import shlex

from click.testing import CliRunner
import pytest

from line_item_manager import cli

@pytest.mark.parametrize("command, echo_str",
 [
  ('--version', 'line-item-manager version'),
  ('', '[OPTIONS] COMMAND [ARGS]'),
 ]
)
def test_version(command, echo_str):
    runner = CliRunner()
    result = runner.invoke(
        cli.cli,
        shlex.split(command)
    )
    assert result.exit_code == 0
    assert echo_str in result.output



from datetime import datetime
import shlex

import pytest

from line_item_manager import cli
from line_item_manager.config import config
from line_item_manager.gam_operations import client as gam_client

def pytest_configure():
    pytest.start_time = datetime(2020, 1, 2, 8, 9, 10)

@pytest.fixture
def cli_config(request):
    config._app = None
    config._cpm_names = None
    config._schema = None

    # patch
    cli_str = request.node.get_closest_marker('command').args[0]
    cli_args = shlex.split(cli_str)
    cmd = cli_args[0]
    ctx = cli.create.make_context(cmd, cli_args[1:])
    if cmd == 'create':
        configfile = ctx.params.pop('configfile')
        config.cli = ctx.params
        config.set_user_configfile(configfile)
        config.set_client_factory(gam_client)
        config.pre_create()
        return dict(
            ctx=ctx,
            configfile=configfile,
            kwargs=ctx.params,
        )
    if cmd == 'show':
        resource = ctx.params.pop('resource')
        return dict(
            ctx=ctx,
            resource=resource,
            kwargs=ctx.params
        )
    return None



"""Top-level package for line-item-manager."""

__author__ = """the prebid contributors"""
__email__ = 'info@prebid.org'
__version__ = '0.2.13'

# For an official release, use dev_version = ''
dev_version = ''

version = __version__
if dev_version:
    version = f'{__version__}-dev{dev_version}'



import shlex

from click.testing import CliRunner
import pytest

from line_item_manager import cli

@pytest.mark.parametrize("command",
 [
  ('config'),
  ('bidders'),
 ]
)
def test_cli_show_good(command):
    runner = CliRunner()
    result = runner.invoke(
        cli.show,
        shlex.split(command)
    )
    assert result.exit_code == 0



#!/usr/bin/env python

"""Tests for `line_item_manager` package."""

import copy
from logging import INFO, WARNING, ERROR
import pytest
import shlex
import yaml
from yaml import safe_dump as dump

from googleads import ad_manager
from googleads.errors import GoogleAdsError, GoogleAdsServerFault
from click.testing import CliRunner

from line_item_manager import cli
from line_item_manager import gam_config
from line_item_manager.config import config, VERBOSE1, VERBOSE2
from line_item_manager.exceptions import ResourceNotActive, ResourceNotFound
from line_item_manager.gam_config import GAMConfig
from line_item_manager.prebid import prebid
from line_item_manager.utils import load_file, num_hash

from .client import MockAdClient, SINGLE_ORDER_SVC_IDS, SINGLE_ORDER_VIDEO_SVC_IDS, \
     BIDDER_BANNER_SVC_IDS, BIDDER_VIDEO_SVC_IDS, BIDDER_TEST_RUN_VIDEO_SVC_IDS, \
     MISSING_RESOURCE_SVC_IDS, BIDDER_BANNER_SVC_IDS_NO_SIZE_OVERRIDE, \
     BIDDER_VIDEO_BIDDER_KEY_MAP_SVC_IDS, BIDDER_VIDEO_SVC_IDS_SIZE_OVERRIDE

CONFIG_FILE = 'tests/resources/cfg.yml'
KEY_FILE = 'tests/resources/gam_creds.json'
CUSTOM_TARGETING = {7101: ['US']}
CONFIG_BIDDER = "interactiveOffers"
if CONFIG_BIDDER not in prebid.bidders:
    prebid.bidders[CONFIG_BIDDER] = {"bidder-code": CONFIG_BIDDER, "bidder-name": "InteractiveOffers"}

EXPECTED_LICA = \
  [[{'creativeId': 4001, 'id': 9001, 'lineItemId': 8001},
    {'creativeId': 4002, 'id': 9002, 'lineItemId': 8001},
    {'creativeId': 4001, 'id': 9001, 'lineItemId': 8002},
    {'creativeId': 4002, 'id': 9002, 'lineItemId': 8002}]]

VIDEO_CREATIVE = {
    'xsi_type': 'VastRedirectCreative',
    'name': 'Prebid InteractiveOffers-video',
    'advertiserId': 1001,
    'size': {'height': 480, 'width': 640},
    'vastXmlUrl': 'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id_interact%%',
    'vastRedirectType': 'LINEAR',
    'duration': 30000
    }

LI_0 = 9999899367
LI_1 = 9999960551

def mock_id(name: str, rec: object):
    return int('9999' + str(num_hash([name, str(rec)])))

CREATIVE_0 = mock_id('CreativeVideo', VIDEO_CREATIVE)
VIDEO_CREATIVE.update({'size': {'height': 240, 'width': 320}})
CREATIVE_1 = mock_id('CreativeVideo', VIDEO_CREATIVE)

LICA_0_0 = mock_id('LICA', {'lineItemId': LI_0, 'creativeId': CREATIVE_0})
LICA_0_1 = mock_id('LICA', {'lineItemId': LI_0, 'creativeId': CREATIVE_1})
LICA_1_0 = mock_id('LICA', {'lineItemId': LI_1, 'creativeId': CREATIVE_0})
LICA_1_1 = mock_id('LICA', {'lineItemId': LI_1, 'creativeId': CREATIVE_1})

DRY_RUN_EXPECTED_LICA = \
  [[{'lineItemId': LI_0, 'creativeId': CREATIVE_0, 'id': LICA_0_0},
    {'lineItemId': LI_0, 'creativeId': CREATIVE_1, 'id': LICA_0_1},
    {'lineItemId': LI_1, 'creativeId': CREATIVE_0, 'id': LICA_1_0},
    {'lineItemId': LI_1, 'creativeId': CREATIVE_1, 'id': LICA_1_1}]]

BANNER_EXPECTED_LICA = \
  [[{'lineItemId': 8001, 'creativeId': 4001, 'sizes': [{'height': 20, 'width': 1000}], 'id': 9001},
    {'lineItemId': 8002, 'creativeId': 4001, 'sizes': [{'height': 20, 'width': 1000}], 'id': 9001}]]

BANNER_EXPECTED_LICA_NO_SIZE_OVERRIDE = \
  [[{'creativeId': 4001, 'id': 9001, 'lineItemId': 8001},
    {'creativeId': 4001, 'id': 9001, 'lineItemId': 8002}]]

VIDEO_EXPECTED_LICA_SIZE_OVERRIDE = \
  [[{'creativeId': 4001,
     'id': 9001,
     'lineItemId': 8001,
     'sizes': [{'height': 480, 'width': 640}, {'height': 240, 'width': 320}]},
    {'creativeId': 4002,
     'id': 9002,
     'lineItemId': 8001,
     'sizes': [{'height': 480, 'width': 640}, {'height': 240, 'width': 320}]},
    {'creativeId': 4001,
     'id': 9001,
     'lineItemId': 8002,
     'sizes': [{'height': 480, 'width': 640}, {'height': 240, 'width': 320}]},
    {'creativeId': 4002,
     'id': 9002,
     'lineItemId': 8002,
     'sizes': [{'height': 480, 'width': 640}, {'height': 240, 'width': 320}]}]]

# init
config._start_time = pytest.start_time
with open(CONFIG_FILE) as fp:
    user = yaml.safe_load(fp)

class Client(MockAdClient):

    def __init__(self, custom_targeting, service_ids, **kwargs):
        self.kwargs = kwargs
        super().__init__(custom_targeting, service_ids)

    def getCustomTargetingKeysByStatement(self, *args):
        r_ = super().getCustomTargetingKeysByStatement(*args)
        if r_ and self.kwargs.get('invalid_targeting_key'):
            r_['results'][0].update(dict(status='INACTIVE'))
        return r_

    def createCustomTargetingValues(self, *args):
        if args[0][0]['customTargetingKeyId'] == 7101:
            assert [rec['name'] for rec in args[0]] == ['CAN']
        return super().createCustomTargetingValues(*args)

    def performOrderAction(self, *args):
        order_ids = [i_['value'] for i_ in args[1]['values'][0]['value']['values']]
        assert order_ids == [6001]
        return dict(numChanges=1)

@pytest.mark.parametrize("command, err_str", [
  (f'tests/resources/cfg_bad_yaml.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   'Check your configfile.'),
  (f'tests/resources/cfg_video.yml -k {KEY_FILE}',
   'You must use --single-order or provide'),
  (f'tests/resources/cfg_video.yml -k {KEY_FILE} -b ix --single-order',
   'Use of --single-order and --bidder-code'),
  (f'tests/resources/cfg_no_pub.yml -k {KEY_FILE} -b ix',
   'Network code must be provided'),
  (f'tests/resources/cfg_no_pub.yml -k {KEY_FILE} -b ix --network-code 1234',
   'Network name must be provided'),
  (f'tests/resources/cfg_video.yml -k {KEY_FILE} -b badcode',
   'Bidder code \'badcode\''),
  (f'tests/resources/cfg_video.yml -k {KEY_FILE} -b ix --network-name badname',
   'Network name found \'Video Publisher\''),
  (f'tests/resources/cfg_validation_error.yml -k {KEY_FILE} -b ix',
   'the following validation errors'),
  (f'tests/resources/cfg_bad_granularity.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   '\'custom\' is a required property'),
  (f'tests/resources/cfg_bad_vcpm.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   '\'vcpm\' requires using line item type \'standard\''),
  (f'tests/resources/cfg_bad_standard_type.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   '\'end_datetime\' is a required property'),
  (f'tests/resources/cfg_bad_timezone.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   'Unknown Time Zone, \'BAD_TIME_ZONE\''),
  (f'tests/resources/cfg_bad_param.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   'Additional properties are not allowed (\'ad_unit_names\' was unexpected)'),
  (f'tests/resources/cfg_bad_key_properties.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   '{\'bad_code\'} must be valid bidder codes'),
  (f'tests/resources/cfg_bad_bidder_keys.yml -k {KEY_FILE} -b {CONFIG_BIDDER}',
   'for \'pubmatic\' must be valid bidder keys'),
])
def test_cli_create_bad(monkeypatch, command, err_str):
    """Test the CLI."""
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)

    runner = CliRunner()
    result = runner.invoke(cli.create, shlex.split(command))
    assert result.exit_code == 2
    assert err_str in result.output

@pytest.mark.parametrize("command, err_str", [
  ('tests/resources/cfg_video.yml -k tests/resources/gam_creds_bad_json.json -b ix', 'Check your private key file.'),
  (f'tests/resources/cfg_video.yml -k {KEY_FILE} -b ix', 'Check your private key file.'),
])
def test_cli_no_mock_create_bad(command, err_str):
    """Test the CLI."""
    runner = CliRunner()
    result = runner.invoke(cli.create, shlex.split(command))
    assert result.exit_code == 2
    assert err_str in result.output

def test_cli_client_exception(monkeypatch):
    command = f'tests/resources/cfg_video.yml -k {KEY_FILE} -b ix'
    def raise_exception(self):
        raise Exception
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", raise_exception)
    runner = CliRunner()
    result = runner.invoke(cli.create, shlex.split(command))
    assert result.exit_code == 2
    assert "Check your private key file. Not able to successfully" in result.output

def test_cli_network_exception(monkeypatch):
    command = f'tests/resources/cfg_video.yml -k {KEY_FILE} -b ix'
    def raise_exception(self):
        raise GoogleAdsError('Test Network')
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    monkeypatch.setattr(gam_config.CurrentNetwork, "fetchone", raise_exception)
    runner = CliRunner()
    result = runner.invoke(cli.create, shlex.split(command))
    assert result.exit_code == 2
    assert "Not able to successfully access your service account" in result.output

@pytest.mark.parametrize("command", [
  (f'tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}'),
])
def test_cli_create_good(monkeypatch, command):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    runner = CliRunner()
    result = runner.invoke(cli.create, shlex.split(command))

    assert result.exit_code == 0

@pytest.mark.parametrize("exc_type, err_str", [
  ('GAE', 'Google Ads Error, Test GAM Error'),
  ('RNF', 'Not able to find the following resource:\n  - Test'),
  ('RNA', 'Resource is not active:\n  - Test'),
  ('KI', 'User Interrupt'),
])
def test_cli_ads_error(caplog, monkeypatch, exc_type, err_str):
    command = f'tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}'
    def raise_exception(self):
        if exc_type == 'GAE':
            raise GoogleAdsError('Test GAM Error')
        if exc_type == 'RNA':
            raise ResourceNotActive('Test')
        if exc_type == 'RNF':
            raise ResourceNotFound('Test')
        if exc_type == 'KI':
            raise KeyboardInterrupt()
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    monkeypatch.setattr(cli.GAMConfig, "create_line_items", raise_exception)
    caplog.set_level(WARNING)
    runner = CliRunner()
    _ = runner.invoke(cli.create, shlex.split(command))

    assert err_str in caplog.text

def test_cli_cleanup_exception(caplog, monkeypatch):
    command = f'tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}'
    def raise_exception(self):
        raise GoogleAdsError('Test GAM Error')
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    monkeypatch.setattr(cli.GAMConfig, "create_line_items", raise_exception)
    monkeypatch.setattr(cli.GAMConfig, "cleanup", raise_exception)
    caplog.set_level(ERROR)
    runner = CliRunner()
    _ = runner.invoke(cli.create, shlex.split(command))

    assert 'Google Ads Error, Test GAM Error' in caplog.text

def test_cli_missing_resource(caplog, monkeypatch):
    command = f'tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}'
    svc_ids = copy.deepcopy(BIDDER_VIDEO_SVC_IDS)
    svc_ids.update(MISSING_RESOURCE_SVC_IDS)
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    caplog.set_level(ERROR)
    runner = CliRunner()
    _ = runner.invoke(cli.create, shlex.split(command))

    assert 'Unexpected result' in caplog.text
    assert 'not found after creation: \'[\'CAN\']\'' in caplog.text

@pytest.mark.command(f'create tests/resources/cfg_video.yml -k {KEY_FILE} --single-order')
def test_video_single_order(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(SINGLE_ORDER_SVC_IDS)
    svc_ids.update(SINGLE_ORDER_VIDEO_SVC_IDS)
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert gam.network['displayName'] == "Video Publisher"
    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_single_order_expected.yml') == \
      gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs
    gam.cleanup()

@pytest.mark.command(f'create tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_archive_error(caplog, monkeypatch, cli_config):
    class ThisClient(Client):
        def performOrderAction(self, *args):
            return dict(numChanges=0)

    client = ThisClient(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    caplog.set_level(ERROR)
    gam = GAMConfig()
    gam.create_line_items()
    gam.cleanup()

    assert 'Order archive, [6001], of 1 changes, reported 0 changes' in caplog.text

@pytest.mark.command(f'create tests/resources/cfg_video.yml -v -v -n -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_dry_run(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()
    gam.cleanup()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected_dry_run.yml') == \
      gam.li_objs[0].line_items
    assert DRY_RUN_EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video.yml -t -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_test_run(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(BIDDER_VIDEO_SVC_IDS)
    svc_ids.update(BIDDER_TEST_RUN_VIDEO_SVC_IDS)
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected_test_run.yml') == \
      gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_one_bidder(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected.yml') == gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video_company_by_id.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_one_bidder(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected.yml') == gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video_one_custom_value.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_one_bidder_custom_criteria(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected_one_custom_value.yml') == gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_one_bidder_invalid_targeting_key(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS, invalid_targeting_key=True)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()

    with pytest.raises(ResourceNotActive) as e_:
        gam.create_line_items()
    assert "'hb_pb_interactiveOff' is not active" in str(e_)

@pytest.mark.command(f'create tests/resources/cfg_video_bidder_key_map.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_one_bidder_key_map(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(BIDDER_VIDEO_SVC_IDS)
    svc_ids.update(BIDDER_VIDEO_BIDDER_KEY_MAP_SVC_IDS)
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected.yml') == gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_banner.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_banner_one_bidder(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_BANNER_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/banner_expected.yml') == gam.li_objs[0].line_items
    assert BANNER_EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video_priority_8.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_priority_8(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert {8} == {i_['priority'] for i_ in gam.li_objs[0].line_items}
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_missing_ad_unit_resource(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(BIDDER_VIDEO_SVC_IDS)
    svc_ids.update(dict(InventoryService={dump(dict(name="ad unit 1")): 2001}))
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()

    with pytest.raises(ResourceNotFound) as e_:
        gam.create_line_items()
    assert "'ad unit 2' was not found" in str(e_)

@pytest.mark.command(f'create tests/resources/cfg_video.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_missing_placement_resource(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(BIDDER_VIDEO_SVC_IDS)
    svc_ids.update(dict(PlacementService={dump(dict(name="placement 2")): 3002}))
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()

    with pytest.raises(ResourceNotFound) as e_:
        gam.create_line_items()
    assert "'placement 1' was not found" in str(e_)

@pytest.mark.command(f'create tests/resources/cfg_video_no_targeting.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_no_targeting_video(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_VIDEO_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected_no_targeting.yml') == \
      gam.li_objs[0].line_items
    assert EXPECTED_LICA == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_banner_vcpm.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_banner_safe_frame_vcpm(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(BIDDER_BANNER_SVC_IDS)
    svc_ids.update(BIDDER_BANNER_SVC_IDS_NO_SIZE_OVERRIDE)
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/banner_vcpm_expected.yml') == gam.li_objs[0].line_items
    assert BANNER_EXPECTED_LICA_NO_SIZE_OVERRIDE == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_video_size_override.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_video_size_override(monkeypatch, cli_config):
    svc_ids = copy.deepcopy(BIDDER_VIDEO_SVC_IDS)
    svc_ids.update(BIDDER_VIDEO_SVC_IDS_SIZE_OVERRIDE)
    client = Client(CUSTOM_TARGETING, svc_ids)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/video_expected.yml') == gam.li_objs[0].line_items
    assert VIDEO_EXPECTED_LICA_SIZE_OVERRIDE == gam.lica_objs

@pytest.mark.command(f'create tests/resources/cfg_sponsorship.yml -k {KEY_FILE} -b {CONFIG_BIDDER}')
def test_sponsorship_priority(monkeypatch, cli_config):
    client = Client(CUSTOM_TARGETING, BIDDER_BANNER_SVC_IDS)
    monkeypatch.setattr(ad_manager.AdManagerClient, "LoadFromString", lambda x: client)
    gam = GAMConfig()
    gam.create_line_items()

    assert len(gam.li_objs) == 1
    assert load_file('tests/resources/sponsorship_expected.yml') == gam.li_objs[0].line_items
    assert BANNER_EXPECTED_LICA == gam.lica_objs



from logging import INFO, WARNING

import pytest
import yaml

from line_item_manager.config import config, VERBOSE1, VERBOSE2
from line_item_manager.prebid import PrebidBidder, prebid
from line_item_manager.utils import package_filename, load_file

CONFIG_FILE = 'tests/resources/cfg.yml'
KEY_FILE = 'tests/resources/gam_creds.json'
TMPL_FILE = 'tests/resources/li_template.yml'
SETTINGS_FILE = 'tests/resources/settings.yml'
SCHEMA_FILE = 'tests/resources/schema.yml'
CONFIG_BIDDER = list(prebid.bidders.keys())[0]

config._start_time = pytest.start_time

with open(CONFIG_FILE) as fp:
    user = yaml.safe_load(fp)

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b interactiveOffers -b ix')
def test_bidders(cli_config):
    assert config.network_code == user['publisher']['network_code']
    assert config.network_name == user['publisher']['network_name']
    assert config.media_types() == ['video', 'banner']
    assert config.start_time == pytest.start_time
    assert not config.isLoggingEnabled(VERBOSE1)
    assert [PrebidBidder(c_).name for c_ in config.bidder_codes()] == \
      ['InteractiveOffers', 'Index Exchange']
    assert config.cpm_names() == ['0.10', '0.20', '0.30', '0.80', '1.30']
    assert [PrebidBidder(c_).targeting_key for c_ in config.bidder_codes()] == \
      ['hb_pb_interactiveOff', 'hb_pb_ix']
    assert config.custom_targeting_key_values() == \
      [{'name': 'country', 'values': {'CAN', 'US'}, 'operator': 'IS', 'reportableType': 'OFF'}]
    assert config.template_src() == open(package_filename('line_item_template.yml')).read()
    assert config.settings_obj() == load_file(package_filename('settings.yml'))
    assert config.schema == load_file(package_filename('schema.yml'))

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} --single-order')
def test_single_order(cli_config):
    assert [PrebidBidder(c_, single_order=config.cli['single_order']).targeting_key \
            for c_ in config.bidder_codes()] == ['hb_pb']
    assert config.user['order']['appliedTeamIds'] == [12345678, 23456789]

def test_fmt_bidder_key():
    assert PrebidBidder(CONFIG_BIDDER).fmt_bidder_key('prefix') == f"prefix_{CONFIG_BIDDER}"
    assert PrebidBidder(CONFIG_BIDDER).fmt_bidder_key('012345678901234') == f"012345678901234_{CONFIG_BIDDER[:4]}"
    assert PrebidBidder('bad', single_order=True).fmt_bidder_key('prefix') == "prefix"

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b {CONFIG_BIDDER} -b ix -t')
def test_test_run(cli_config):
    assert config.cpm_names() == ['0.10', '0.20']
    assert config.user['creative']['video']['duration'] == 30000
    assert config.user['creative']['video']['max_duration'] == 30000

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b {CONFIG_BIDDER} -b ix --template {TMPL_FILE}')
def test_template(cli_config):
    assert config.template_src() == open(TMPL_FILE).read()

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b {CONFIG_BIDDER} -b ix --settings {SETTINGS_FILE}')
def test_settings(cli_config):
    assert config.settings_obj() == load_file(SETTINGS_FILE)

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b {CONFIG_BIDDER} -b ix --schema {SCHEMA_FILE}')
def test_schema(cli_config):
    assert config.schema == load_file(SCHEMA_FILE)

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} --network-code 9876 --network-name abcd -b {CONFIG_BIDDER} -b ix')
def test_network_meta(cli_config):
    assert config.network_code == 9876
    assert config.network_name == 'abcd'

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b ix -v')
def test_verbose1(cli_config):
    assert config.isLoggingEnabled(VERBOSE1)
    assert not config.isLoggingEnabled(VERBOSE2)

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b ix -v -v')
def test_verbose2(cli_config):
    assert config.isLoggingEnabled(VERBOSE1)
    assert config.isLoggingEnabled(VERBOSE2)

@pytest.mark.command(f'create {CONFIG_FILE} -k {KEY_FILE} -b ix -q')
def test_quiet(cli_config):
    assert config.isLoggingEnabled(WARNING)
    assert not config.isLoggingEnabled(INFO)

@pytest.mark.command(f'create tests/resources/cfg_granularity_high.yml -k {KEY_FILE} -b ix -q')
def test_granularity(cli_config):
    assert config.cpm_names() == ['%.2f' % (v_ / 100) for v_ in range(1, 2001, 1)]

@pytest.mark.command(f'create tests/resources/cfg_is_not_operator.yml -k {KEY_FILE} -b ix -q')
def test_custom_targeting_is_not_reportableType(cli_config):
    assert config.custom_targeting_key_values() == \
      [{'name': 'country', 'values': {'CAN', 'US'}, 'operator': 'IS_NOT', 'reportableType': 'ON'}]
    assert config.targeting_bidder_key_config() == {'reportableType': 'ON'}

@pytest.mark.command(f'create tests/resources/cfg_video_max_duration.yml -k {KEY_FILE} -b ix -q')
def test_video_max_duration(cli_config):
    assert config.user['creative']['video']['duration'] == 60000
    assert config.user['creative']['video']['max_duration'] == 60000

@pytest.mark.command(f'create tests/resources/cfg_video_duration.yml -k {KEY_FILE} -b ix -q')
def test_video_duration(cli_config):
    assert config.user['creative']['video']['duration'] == 15000
    assert config.user['creative']['video']['max_duration'] == 30000



#!/usr/bin/env python

"""Tests for `line_item_manager` package."""

from datetime import datetime, timezone
import pytest

from line_item_manager.utils import ichunk, format_long_list
from line_item_manager.yaml_date import date_from_string


def test_ichunk():
    assert list(ichunk(range(10), 3)) == [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

def test_format_long_list():
    assert format_long_list([f'abc_{i_}' for i_ in range(10)]) == \
      "['abc_0', 'abc_1', 'abc_2', ..., 'abc_7', 'abc_8', 'abc_9']"
    assert format_long_list([f'abc_{i_}' for i_ in range(4)]) == \
      "['abc_0', 'abc_1', 'abc_2', 'abc_3']"

def test_date_from_string():
    assert date_from_string('11/21/20 12:11', '%m/%d/%y %H:%M', 'UTC') == \
      datetime(2020, 11, 21, 12, 11, tzinfo=timezone.utc)



import copy
from yaml import safe_dump as dump, safe_load as loads

SVC_IDS = dict(
    NetworkService={
        dump(dict()): 1501,
    },
    CompanyService={
        dump(dict(id=1001)): 1001,
        dump(dict(name="Prebid-InteractiveOffers")): 1001,
    },
    InventoryService={
        dump(dict(name="ad unit 1")): 2001,
        dump(dict(name="ad unit 2")): 2002,
    },
    PlacementService={
        dump(dict(name="placement 1")): 3001,
        dump(dict(name="placement 2")): 3002,
    },
    UserService={
        dump(dict()): 5001,
    },
    CustomTargetingService={
        dump(dict(name="country")): 7101,
        dump(dict(customTargetingKeyId=7101, name="US")): 7301,
        dump(dict(customTargetingKeyId=7101, name="CAN")): 7302,
        dump(dict(name="hb_pb_interactiveOff")): 7201,
        dump(dict(customTargetingKeyId=7201, name="1.25")): 7401,
        dump(dict(customTargetingKeyId=7201, name="1.50")): 7402,
    },
    LineItemCreativeAssociationService={
        dump(dict(lineItemId=8001, creativeId=4001)): 9001,
        dump(dict(lineItemId=8001, creativeId=4002)): 9002,
        dump(dict(lineItemId=8002, creativeId=4001)): 9001,
        dump(dict(lineItemId=8002, creativeId=4002)): 9002,
    },
)

SINGLE_ORDER_SVC_IDS = dict(
    CompanyService={
        dump(dict(name="Prebid-Top Bid")): 1001,
    },
    CustomTargetingService={
        dump(dict(name="country")): 7101,
        dump(dict(customTargetingKeyId=7101, name="US")): 7301,
        dump(dict(customTargetingKeyId=7101, name="CAN")): 7302,
        dump(dict(name="hb_pb")): 7501,
        dump(dict(customTargetingKeyId=7501, name="1.25")): 7601,
        dump(dict(customTargetingKeyId=7501, name="1.50")): 7602,
    },
)

MISSING_RESOURCE_SVC_IDS = dict(
    CustomTargetingService={
        dump(dict(name="country")): 7101,
        dump(dict(customTargetingKeyId=7101, name="US")): 7301,
        dump(dict(name="hb_pb")): 7501,
        dump(dict(customTargetingKeyId=7501, name="1.25")): 7601,
        dump(dict(customTargetingKeyId=7501, name="1.50")): 7602,
    },
)

SINGLE_ORDER_VIDEO_SVC_IDS = dict(
    CreativeService={
        dump(dict(
            name="Prebid Top Bid-video",
            advertiserId=1001,
            size={'height': 480, 'width': 640},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id%%',
        )): 4001,
        dump(dict(
            name="Prebid Top Bid-video",
            advertiserId=1001,
            size={'height': 240, 'width': 320},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id%%',
        )): 4002,
    },
    OrderService={
        dump(dict(
            name="Prebid-Top Bid-video-01/02/2020-08:09:10 1.25-1.50",
            advertiserId=1001,
            traffickerId=5001,
        )): 6001,
    },
    LineItemService={
        dump(dict(
            name="Prebid-Top Bid-video-01/02/2020-08:09:10 @ 1.25",
            orderId=6001,
        )): 8001,
        dump(dict(
            name="Prebid-Top Bid-video-01/02/2020-08:09:10 @ 1.50",
            orderId=6001,
        )): 8002,
    },
)

BIDDER_VIDEO_SVC_IDS = dict(
    CreativeService={
        dump(dict(
            name="Prebid InteractiveOffers-video",
            advertiserId=1001,
            size={'height': 480, 'width': 640},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id_interact%%',
        )): 4001,
        dump(dict(
            name="Prebid InteractiveOffers-video",
            advertiserId=1001,
            size={'height': 240, 'width': 320},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id_interact%%',
        )): 4002,
    },
    OrderService={
        dump(dict(
            name="Prebid-InteractiveOffers-video-01/02/2020-08:09:10 1.25-1.50",
            advertiserId=1001,
            traffickerId=5001,
        )): 6001,
    },
    LineItemService={
        dump(dict(
            name="Prebid-InteractiveOffers-video-01/02/2020-08:09:10 @ 1.25",
            orderId=6001,
        )): 8001,
        dump(dict(
            name="Prebid-InteractiveOffers-video-01/02/2020-08:09:10 @ 1.50",
            orderId=6001,
        )): 8002,
    },
)

BIDDER_VIDEO_SVC_IDS_SIZE_OVERRIDE = dict(
    CreativeService={
        dump(dict(
            name="Prebid InteractiveOffers-video Copy:1",
            advertiserId=1001,
            size={'height': 1, 'width': 1},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id_interact%%',
        )): 4001,
        dump(dict(
            name="Prebid InteractiveOffers-video Copy:2",
            advertiserId=1001,
            size={'height': 1, 'width': 1},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:hb_cache_id_interact%%',
        )): 4002,
    },
)

BIDDER_TEST_RUN_VIDEO_SVC_IDS = dict(
    OrderService={
        dump(dict(
            name="Test: Prebid-InteractiveOffers-video-01/02/2020-08:09:10 1.25-1.50",
            advertiserId=1001,
            traffickerId=5001,
        )): 6001,
    },
    LineItemService={
        dump(dict(
            name="Test: Prebid-InteractiveOffers-video-01/02/2020-08:09:10 @ 1.25",
            orderId=6001,
        )): 8001,
        dump(dict(
            name="Test: Prebid-InteractiveOffers-video-01/02/2020-08:09:10 @ 1.50",
            orderId=6001,
        )): 8002,
    },
)

BIDDER_BANNER_SVC_IDS = dict(
    CreativeService={
        dump(dict(
            name="Prebid InteractiveOffers-banner Copy:1",
            advertiserId=1001,
            size={'height': 1, 'width': 1},
        )): 4001,
    },
    OrderService={
        dump(dict(
            name="Prebid-InteractiveOffers-banner-01/02/2020-08:09:10 1.25-1.50",
            advertiserId=1001,
            traffickerId=5001,
        )): 6001,
    },
    LineItemService={
        dump(dict(
            name="Prebid-InteractiveOffers-banner-01/02/2020-08:09:10 @ 1.25",
            orderId=6001,
        )): 8001,
        dump(dict(
            name="Prebid-InteractiveOffers-banner-01/02/2020-08:09:10 @ 1.50",
            orderId=6001,
        )): 8002,
    },
)

BIDDER_BANNER_SVC_IDS_NO_SIZE_OVERRIDE = dict(
    CreativeService={
        dump(dict(
            name="Prebid InteractiveOffers-banner",
            advertiserId=1001,
            size={'height': 20, 'width': 1000},
        )): 4001,
    },
)

BIDDER_VIDEO_BIDDER_KEY_MAP_SVC_IDS = dict(
    CustomTargetingService={
        dump(dict(name="country")): 7101,
        dump(dict(customTargetingKeyId=7101, name="US")): 7301,
        dump(dict(customTargetingKeyId=7101, name="CAN")): 7302,
        dump(dict(name="io_custom_key")): 7201,
        dump(dict(customTargetingKeyId=7201, name="1.25")): 7401,
        dump(dict(customTargetingKeyId=7201, name="1.50")): 7402,
    },
    CreativeService={
        dump(dict(
            name="Prebid InteractiveOffers-video",
            advertiserId=1001,
            size={'height': 480, 'width': 640},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:io_custom_cache_id%%',
        )): 4001,
        dump(dict(
            name="Prebid InteractiveOffers-video",
            advertiserId=1001,
            size={'height': 240, 'width': 320},
            vastXmlUrl= \
            'https://prebid.adnxs.com/pbc/v1/cache?uuid=%%PATTERN:io_custom_cache_id%%',
        )): 4002,
    },
)

def rec_from_statement(statement):
    rec = {}
    for i_ in statement['values']:
        if i_['value']['xsi_type'] == 'SetValue':
            value = [v['value'] for v in i_['value']['values']]
        else:
            value = i_['value']['value']
        rec[i_['key']] = value
    return rec

def svc_id(svc_recs, rec):
    _id = None
    for key, val in svc_recs.items():
        if loads(key).items() <= rec.items():
            assert _id is None
            _id = val
    return _id

def byStatement(self, *args):
    rec = rec_from_statement(args[0])
    _id = svc_id(self.svc_ids[self.service], rec)
    if not _id:
        return {}
    rec.update({'id': _id})
    return dict(results=[rec])

def create(self, *args):
    recs = copy.deepcopy(args[0])
    out = []
    for rec in recs:
        id_ = svc_id(self.svc_ids[self.service], rec)
        if id_:
            rec.update({'id': id_})
            out.append(rec)
    return out

class MockAdClient:

    def __init__(self, custom_targeting, service_ids):
        self.custom_targeting = copy.deepcopy(custom_targeting)
        self.svc_ids = copy.deepcopy(SVC_IDS)
        self.svc_ids.update(service_ids)

    def GetService(self, service, version=None):
        self.service = service
        return self

    def getCurrentUser(self, *args):
        return dict(id=svc_id(self.svc_ids[self.service], dict()))

    def getCurrentNetwork(self, *args):
        return dict(
            id=svc_id(self.svc_ids[self.service], dict()),
            displayName="Video Publisher",
            effectiveRootAdUnitId=1511,
        )

    def getCreativesByStatement(self, *args):
        return {}

    def getCustomTargetingValuesByStatement(self, *args):
        rec = rec_from_statement(args[0])
        recs = []
        for val in self.custom_targeting.get(rec['customTargetingKeyId'], []):
            if val in set(rec['name']):
                r_ = copy.deepcopy(rec)
                r_.update(dict(name=val))
                r_.update({'id': svc_id(self.svc_ids[self.service], r_)})
                recs.append(r_)
        return dict(results=recs)

for i_ in ('AdUnits', 'Placements', 'Companies', 'Orders', 'CustomTargetingKeys'):
    setattr(MockAdClient, f'get{i_}ByStatement', byStatement)

for i_ in ('Creatives', 'LineItems', 'LineItemCreativeAssociations', 'CustomTargetingValues'):
    setattr(MockAdClient, f'create{i_}', create)



#!/usr/bin/env python

"""Tests for `line_item_manager` package."""

import yaml

from line_item_manager.yaml_date import date_from_string
from jinja2 import Template as J2Template

def test_yaml_tz():
    dt = date_from_string('11/21/20 12:11', '%m/%d/%y %H:%M', 'Europe/Oslo')
    obj = yaml.safe_load(yaml.safe_dump({'dt': dt}))
    assert obj['dt'].tzinfo.zone == 'Europe/Oslo'

def test_jinja_tz():
    dt = date_from_string('11/21/20 12:11', '%m/%d/%y %H:%M', 'Europe/Oslo')
    obj = yaml.safe_load(J2Template("dt: {{ dt }}").render(dt=dt))
    assert obj['dt'].tzinfo.zone == 'Europe/Oslo'



from datetime import datetime
from pprint import pformat
from typing import Any, List, Iterable, Optional

from googleads.errors import GoogleAdsServerFault
from jinja2 import Template as J2Template
from retrying import retry
from tqdm import tqdm
from zeep.helpers import serialize_object
import pytz

from .config import config, VERBOSE1, VERBOSE2
from .exceptions import ResourceNotActive, ResourceNotFound
from .operations import Advertiser, AdUnit, Placement, TargetingKey, TargetingValues, \
     CreativeBanner, CreativeVideo, Order, CurrentNetwork, CurrentUser, LineItem, LICA
from .prebid import PrebidBidder
from .template import render_cfg, render_src
from .utils import format_long_list, ichunk

logger = config.getLogger(__name__)

def is_create_retryable_error(exc: Exception) -> bool:
    is_error = isinstance(exc, GoogleAdsServerFault)
    if is_error:
        logger.info('Retrying...')
    return is_error

def log(objname: str, obj: dict=None) -> None:
    logger.log(VERBOSE1, '%s:\n%s', objname, pformat(obj if obj else config.user.get(objname, {})))

def target_fetch(key: str, names: Iterable[str], operator='IS', match_type: str='EXACT',
                 reportableType: str='OFF') -> dict:
    tgt_key = TargetingKey(name=key, reportableType=reportableType).fetchone(create=True)
    recs = []
    for name in names:
        recs.append(dict(
            customTargetingKeyId=tgt_key['id'],
            name=name,
            displayName=name,
            matchType=match_type,
        ))
    tgt_values = TargetingValues(key_id=tgt_key['id'], name=list(names)).fetch(create=True, recs=recs, validate=True)
    return dict(
        key=tgt_key,
        operator=operator,
        values=tgt_values,
        names={v['name']:v for v in tgt_values}
    )

class GAMLineItems:

    def __init__(self, gam: Any, media_type: str, bidder: PrebidBidder, cpms: List[str]):
        self._advertiser: Optional[dict] = None
        self._creatives: Optional[List[dict]] = None
        self._line_items: Optional[List[dict]] = None
        self._order: Optional[dict] = None
        self._targeting_key: Optional[dict] = None

        self.gam: GAMConfig = gam
        self.media_type = media_type
        self.bidder = bidder
        self.cpms = cpms
        self.atts = dict(
            bidder=bidder,
            media_type=media_type,
        )
        _ = self.advertiser

    @property
    def is_size_override(self) -> bool:
        default = config.app['mgr']['creative'][self.media_type]['size_override']
        return config.user['creative'][self.media_type].get('size_override', default)

    @property
    def advertiser(self) -> dict:
        if self._advertiser is None:
            cfg = render_cfg('advertiser', self.bidder)
            log('advertiser', obj=cfg)
            if cfg.get('id'):
                self._advertiser = Advertiser(id=cfg['id']).fetchone()
                if not 'id' in self._advertiser:
                    raise ResourceNotFound(f"Advertiser id: {cfg['id']} was not found")
            else:
                self._advertiser = \
                  Advertiser(name=cfg['name'],
                             type=cfg.get('type', config.app['mgr']['advertiser']['type'])).fetchone(create=True)
        return self._advertiser

    @retry(retry_on_exception=is_create_retryable_error, stop_max_attempt_number=5, wait_fixed=2000)
    def create_licas(self, recs: List[dict]) -> List[dict]:
        return LICA().create(recs, validate=True)

    def create_licas_batched(self, recs: List[dict]) -> List[dict]:
        out = []
        logger.info(f'Line Item Creative Associations: Writing {len(recs)} records...')
        with tqdm(total=len(recs)) as pbar:
            for subset in ichunk(recs, config.app['mgr']['max_lica_records']):
                out += self.create_licas(subset)
                pbar.update(len(subset))
        return out

    def create(self) -> List[dict]:
        recs = []
        for line_item in self.line_items:
            for creative in self.creatives:
                rec = dict(lineItemId=line_item['id'], creativeId=creative['id'])
                if self.is_size_override:
                    rec.update(dict(sizes=config.user['creative'][self.media_type]['sizes']))
                recs.append(rec)
        return self.create_licas_batched(recs)

    @property
    def creatives(self) -> List[dict]:
        if self._creatives is None:
            cfg = render_cfg('creative', self.bidder, media_type=self.media_type)
            _name = f'creative_{self.media_type}'
            _method = getattr(self, _name)
            log(_name, obj={k:cfg[k] for k in ('name', self.media_type)})
            self._creatives = [_method(i_, cfg, size) \
                               for i_, size in enumerate(cfg[self.media_type]['sizes'])]
        return self._creatives

    def creative_name(self, cfg: dict, index: int):
        if self.is_size_override:
            tmpl = config.app['mgr']['creative']['size_override']['name_template']
            return J2Template(tmpl).render(name=cfg['name'], index=index + 1)
        return cfg['name']

    def creative_banner(self, index: int, cfg: dict, size: dict) -> dict:
        params = dict(
            name=self.creative_name(cfg, index),
            advertiserId=self.advertiser['id'],
            size=config.app['prebid']['creative']['size_override'] if self.is_size_override else size,
            snippet=cfg['banner']['snippet'],
            isSafeFrameCompatible=cfg['banner'].get('safe_frame', True),
        )
        return CreativeBanner(**params).fetchone(create=True)

    def creative_video(self, index: int, cfg: dict, size: dict) -> dict:
        params = dict(
            name=self.creative_name(cfg, index),
            advertiserId=self.advertiser['id'],
            size=config.app['prebid']['creative']['size_override'] if self.is_size_override else size,
            vastXmlUrl=cfg['video']['vast_xml_url'],
            duration=config.user['creative']['video']['duration'],
        )
        return CreativeVideo(**params).fetchone(create=True)

    @property
    def line_items(self) -> List[dict]:
        if self._line_items is None:
            recs = []
            src = config.template_src()
            for i_, cpm in enumerate(self.cpms):
                li_cfg = render_cfg('line_item', self.bidder, cpm=cpm, media_type=self.media_type)
                if (i_ == 0) or (i_ == len(self.cpms) - 1) or config.isLoggingEnabled(VERBOSE2):
                    log('line_item', obj=li_cfg)
                params = dict(
                    micro_amount=config.micro_amount(cpm),
                    cpm=cpm,
                    li=self,
                    li_cfg=li_cfg,
                    user_cfg=config.user,
                )
                recs.append(render_src(src, **params))
            self._line_items = LineItem().create(recs, validate=True)
        return self._line_items

    @property
    def order(self) -> dict:
        if self._order is None:
            cfg = render_cfg('order', self.bidder, media_type=self.media_type,
                             cpm_min=self.cpms[0], cpm_max=self.cpms[-1])
            log('order', obj=cfg)
            self._order = Order(
                name=cfg['name'],
                advertiserId=self.advertiser['id'],
                traffickerId=self.gam.user['id'],
                appliedTeamIds=config.user['order'].get('appliedTeamIds'),
            ).fetchone(create=True)
        return self._order

    @property
    def targeting_key(self) -> dict:
        if self._targeting_key is None:
            reportableType=config.targeting_bidder_key_config().get('reportableType', 'OFF')
            self._targeting_key = target_fetch(
                self.bidder.targeting_key, config.cpm_names(), reportableType=reportableType)
        return self._targeting_key

class GAMConfig:

    def __init__(self):
        _ = [log(i_) for i_ in ('targeting', 'rate')]
        self._ad_units: Optional[List[dict]] = None
        self._bidders: Optional[List[PrebidBidder]] = None
        self._li_objs: List[GAMLineItems] = []
        self._lica_objs: List[List[dict]] = []
        self._network: Optional[dict] = None
        self._placements: Optional[List[dict]] = None
        self._targeting_custom: Optional[List[dict]] = None
        self._user: Optional[dict] = None

        self._success = False

    @property
    def li_objs(self) -> List[GAMLineItems]:
        return self._li_objs

    @property
    def lica_objs(self) -> List[List[dict]]:
        return self._lica_objs

    @property
    def ad_units(self) -> List[dict]:
        if self._ad_units is None:
            self._ad_units = []
            for name in config.user.get('targeting', {}).get('ad_unit_names', []):
                ad_unit = AdUnit(name=name).fetchone()
                if not ad_unit:
                    raise ResourceNotFound(f'Ad Unit named \'{name}\' was not found')
                self._ad_units.append(ad_unit)
        return self._ad_units

    def add_li_obj(self, media_type: str, bidder: PrebidBidder, cpms: List[str]) -> GAMLineItems:
        self._li_objs.append(GAMLineItems(self, media_type, bidder, cpms))
        return self._li_objs[-1]

    def archive(self) -> None:
        order_ids = [i_.order['id'] for i_ in self._li_objs]
        if order_ids:
            logger.info('Auto-archiving Orders:\n%s', pformat(order_ids))
            response = Order(id=order_ids).archive()
            changes = response['numChanges'] if 'numChanges' in response else None
            if not changes == len(order_ids):
                logger.error('Order archive, %s, of %d changes, reported %s changes',
                             order_ids, len(order_ids), changes)

    @property
    def bidders(self) -> List[PrebidBidder]:
        if self._bidders is None:
            self._bidders = []
            for code in config.bidder_codes():
                bidder = PrebidBidder(
                    code,
                    override_map=config.user.get('bidder_key_map', {}).get(code, {}),
                    single_order=config.cli['single_order']
                    )
                tgt_key = serialize_object(TargetingKey(name=bidder.targeting_key).fetchone())
                if tgt_key and not tgt_key.get('status', 'ACTIVE') == 'ACTIVE':
                    raise ResourceNotActive(f"Bidder Targeting Key name \'{tgt_key['name']}\' "
                                            "is not active.  You must activate before re-running.")
                self._bidders.append(bidder)
        return self._bidders

    def cleanup(self) -> None:
        if not self.success and not config.cli['skip_auto_archive']:
            self.archive()

    def check_resources(self) -> None:
        _ = self.ad_units
        _ = self.placements
        _ = self.bidders

    def create_line_items(self) -> None:
        self.check_resources()
        for bidder in self.bidders:
            logger.info('#' * 80)
            logger.info('Bidder: name="%s", code="%s"', bidder.name, bidder.code)
            logger.info('Key: "%s", Values: %s', bidder.targeting_key,
                        format_long_list(config.cpm_names()))
            for media_type in config.media_types():
                logger.info('#' * 60)
                logger.info('Media Type: "%s"', media_type)
                for cpms in config.cpm_names_batched():
                    logger.info('Line Items: CPMs(min=%s, max=%s, count=%d)',
                                cpms[0], cpms[-1], len(cpms))
                    li_ = self.add_li_obj(media_type, bidder, cpms)
                    logger.info('Line Item Creative Associations: Creative Count=%d',
                                len(li_.creatives))
                    self._lica_objs.append(li_.create())

    @property
    def network(self) -> dict:
        if self._network is None:
            self._network = CurrentNetwork().fetchone()
        return self._network

    @property
    def placements(self) -> List[dict]:
        if self._placements is None:
            self._placements = []
            for name in config.user.get('targeting', {}).get('placement_names', []):
                placement = Placement(name=name).fetchone()
                if not placement:
                    raise ResourceNotFound(f'Placement named \'{name}\' was not found')
                self._placements.append(placement)
        return self._placements

    @property
    def success(self) -> bool:
        return self._success

    @success.setter
    def success(self, val: bool) -> None:
        self._success = val

    @property
    def targeting_custom(self) -> List[dict]:
        if self._targeting_custom is None:
            self._targeting_custom = [target_fetch(
                D_['name'],
                D_['values'],
                operator=D_['operator'],
                reportableType=D_['reportableType']) \
                for D_ in config.custom_targeting_key_values()]
        return self._targeting_custom

    @property
    def user(self) -> dict:
        if self._user is None:
            self._user = CurrentUser().fetchone()
        return self._user



class ResourceNotActive(Exception):
    """
    Raised when specified resource is not active
    """

class ResourceNotFound(Exception):
    """
    Raised when specified resource is not found
    """



from datetime import datetime
from typing import Optional
import pytz
import yaml

YAML_TAG = u'!datetime_tz'

class DateTimeTZ(datetime):
    def __new__(cls, *args, **kwargs):
        return datetime.__new__(cls, *args, **kwargs)
    def __str__(self, include_tag=True):
        """Represent time zone name for encoding which is needed by googleads"""
        tag = f'{YAML_TAG} ' if include_tag else ''
        return f'{tag}{super().__str__()};{self.tzinfo.zone}'

def tz_constructor(loader: yaml.loader.SafeLoader, node: yaml.nodes.ScalarNode) -> datetime:
    dt_str, timezone = loader.construct_scalar(node).split(';')
    return datetime.fromisoformat(dt_str).replace(tzinfo=pytz.timezone(timezone))

def tz_representer(dumper: yaml.dumper.SafeDumper, data: DateTimeTZ) -> yaml.nodes.ScalarNode:
    return dumper.represent_scalar(YAML_TAG, f'{data.__str__(include_tag=False)}')

yaml.representer.SafeRepresenter.add_representer(DateTimeTZ, tz_representer)
yaml.constructor.SafeConstructor.add_constructor(YAML_TAG, tz_constructor)

def date_from_string(dtstr: str, fmt: str, timezone: str) -> Optional[DateTimeTZ]:
    """Get datetime object from a date string.

    Args:
      dtstr: input date string
      fmt: datetime format string
      timezone: time zone string

    Returns:
      A DateTimeTZ object if date string is provide else None
    """
    if not dtstr:
        return None
    dt = datetime.strptime(dtstr, fmt)
    return DateTimeTZ(dt.year, dt.month, dt.day, dt.hour, dt.minute).\
      replace(tzinfo=pytz.timezone(timezone))



import copy
from typing import List, Optional, Tuple, Union

from googleads import ad_manager

from .config import config
from .gam_operations import GAMOperations
from .utils import num_hash

class AppOperations(GAMOperations):
    """Operations using GAM abstract class."""

    @property
    def client(self) -> Optional[ad_manager.AdManagerClient]:
        """Get GAM client.

        Returns:
          GAM client
        """
        return config.client

    @property
    def version(self) -> str:
        """Get preferred GAM API version

        Returns:
           version string
        """
        return config.app['googleads']['version']

    @property
    def dry_run(self) -> bool:
        """Get dry run state.

        Returns:
          True if doing a dry run
        """
        return config.cli.get('dry_run', False)

    def create_id(self, rec: dict) -> int:
        """Create ID from reference rec.

        Args:
          rec: reference

        Returns:
          Integer hash as ID
        """
        n_ = num_hash([type(self).__name__, str(rec)])
        return int(''.join([str(config.app['mgr']['dry_run']['id_prefix']), str(n_)]))

    def dry_run_recs(self, recs: List[dict]) -> List[dict]:
        """Recs that are returned on create when doing a dry run.

        Args:
          recs: reference recs

        Returns:
          Reference recs with dummy IDs
        """
        out = copy.deepcopy(recs)
        _ = [r_.update(dict(id=self.create_id(r_))) for r_ in out]
        return out

    def check(self, rec: dict) -> Union[str, Tuple[int, int]]:
        """Record attribute used for existence checks.

        Args:
          rec: reference rec

        Returns:
          Record attribute
        """
        return rec['name']

    def validate(self, recs: List[dict], results: List[dict]) -> None:
        """Raise value error if records are missing.

        Args:
          recs: reference recs
          results: operation results
        """
        observed = {self.check(r_) for r_ in results}
        missing = [self.check(r_) for r_ in recs if self.check(r_) not in observed]
        if missing:
            raise ValueError(f'Following items were not found after creation: \'{missing}\'')

class AdUnit(AppOperations):
    service = 'InventoryService'
    method = 'getAdUnitsByStatement'

class Advertiser(AppOperations):
    service = 'CompanyService'
    method = 'getCompaniesByStatement'
    create_method = 'createCompanies'

    def __init__(self, *args, **kwargs):
        if 'type' in kwargs:
            kwargs['type'] = kwargs['type']
        super().__init__(*args, **kwargs)

class Creative(AppOperations):
    service = 'CreativeService'
    method = 'getCreativesByStatement'
    create_method = 'createCreatives'
    query_fields = ('id', 'name', 'advertiserId', 'width', 'height')

    def __init__(self, *args, **kwargs):
        if 'size' in kwargs:
            kwargs['height'] = kwargs['size']['height']
            kwargs['width'] = kwargs['size']['width']
        super().__init__(*args, **kwargs)

class CreativeVideo(Creative):
    create_fields = ('xsi_type', 'name', 'advertiserId', 'size', 'vastXmlUrl',
                     'vastRedirectType', 'duration')

    def __init__(self, *args, xsi_type: str='VastRedirectCreative', vastRedirectType: str='LINEAR',
                 duration: int, **kwargs):
        kwargs['xsi_type'] = xsi_type
        kwargs['vastRedirectType'] = vastRedirectType
        kwargs['duration'] = duration if duration else \
          config.app['prebid']['creative']['video']['max_duration']
        super().__init__(*args, **kwargs)

class CreativeBanner(Creative):
    create_fields = ('xsi_type', 'name', 'advertiserId', 'size', 'isSafeFrameCompatible', 'snippet')

    def __init__(self, *args, xsi_type: str='ThirdPartyCreative',
                 isSafeFrameCompatible: bool=True, **kwargs):
        kwargs['xsi_type'] = xsi_type
        kwargs['isSafeFrameCompatible'] = isSafeFrameCompatible
        super().__init__(*args, **kwargs)

class CurrentNetwork(AppOperations):
    use_statement = False
    service = 'NetworkService'
    method = 'getCurrentNetwork'

class CurrentUser(AppOperations):
    use_statement = False
    service = 'UserService'
    method = 'getCurrentUser'

class LICA(AppOperations):
    service = 'LineItemCreativeAssociationService'
    create_method = 'createLineItemCreativeAssociations'
    method = 'getLineItemCreativeAssociationsByStatement'

    def check(self, rec: dict) -> Tuple[int, int]:
        return (rec['lineItemId'], rec['creativeId'])

class LineItem(AppOperations):
    service = 'LineItemService'
    method = 'getLineItemsByStatement'
    create_method = 'createLineItems'

class Order(AppOperations):
    service = 'OrderService'
    method = 'getOrdersByStatement'
    create_method = 'createOrders'
    query_fields = ('id', 'name', 'advertiserId', 'traffickerId')

    def __init__(self, *args, **kwargs):
        if 'appliedTeamIds' in kwargs and kwargs['appliedTeamIds'] is None:
            del kwargs['appliedTeamIds']
        super().__init__(*args, **kwargs)

    def archive(self) -> dict:
        if self.dry_run:
            return dict(numChanges=len(self.params['id']))
        return self.svc().performOrderAction(
            {'xsi_type': 'ArchiveOrders'},
            self.statement().ToStatement()) # type: ignore[union-attr]

class Placement(AppOperations):
    service = 'PlacementService'
    method = 'getPlacementsByStatement'

class TargetingKey(AppOperations):
    service = 'CustomTargetingService'
    method = 'getCustomTargetingKeysByStatement'
    create_method = 'createCustomTargetingKeys'
    query_fields = ('name', )

    def __init__(self, *args, name: str=None, _type: str='PREDEFINED', **kwargs):
        if not name is None:
            kwargs['name'] = name
        kwargs['displayName'] = kwargs.get('displayName', name)
        kwargs['type'] = _type
        super().__init__(*args, **kwargs)

class TargetingValues(AppOperations):
    service = 'CustomTargetingService'
    method = 'getCustomTargetingValuesByStatement'
    create_method = 'createCustomTargetingValues'

    def __init__(self, *args, key_id: int=None, **kwargs):
        kwargs['customTargetingKeyId'] = key_id
        super().__init__(*args, **kwargs)



from hashlib import sha1
import pkg_resources
from pprint import pformat
from typing import Any, Dict, Iterable, List

import yaml

def load_file(filename: str) -> dict:
    """Load a yaml file returning the represented dict.

    Args:
      filename: full path of file

    Returns:
      A dict based on specified yaml file
    """
    with open(filename) as fp:
        return yaml.safe_load(fp)

def load_package_file(name: str) -> dict:
    """Load a yaml package file returning the represented dict.

    Args:
      name: base name of the package file

    Returns:
      A dict based on specified yaml packaged file
    """
    return load_file(package_filename(name))

def package_filename(name: str) -> str:
    """Get fullpath of a package filename specified by the base filename.

    Args:
      name: base name of the package file

    Returns:
      Fullpath of the package file
    """
    return pkg_resources.resource_filename('line_item_manager',
                                           f'conf.d/{name}') # type: ignore[misc]

def read_package_file(name: str) -> str:
    """Get contents of a package file specified by name.

    Args:
      name: base name of the package file

    Returns:
      Contents of the package file as a string
    """
    with open(package_filename(name)) as fp:
        return fp.read()

def num_hash(obj: Any, digits: int=6) -> int:
    """Get an integer hash with specified number of digits.

    Args:
      obj: object to hash
      digits: number of digits in returned hash

    Returns:
      A hashed integer with specified number of digits
    """
    return int(sha1(str(obj).encode('utf-8')).hexdigest(), 16) % 10**digits

def values_from_bucket(bucket: Dict[str, float]) -> set:
    """Get set of price formatted values specified by min, max and interval.

    Args:
      bucket: dict containing min, max and interval values

    Returns:
      Formatted set of values from min to max by interval
    """
    rng = [round(100 * bucket[_k]) for _k in ('min', 'max', 'interval')]
    rng[1] += rng[2] # make stop inclusive
    return {_x / 100 for _x in range(*rng)}

def format_long_list(vals: list, cnt: int=3) -> str:
    """Pretty format with head, tail, and ellipsis to indicate omissions.

    Args:
      vals: sequence of values to be formatted
      cnt: size of the head and tail

    Returns:
      A pretty formatted string with omissions
    """
    fmt = pformat(vals)
    out = fmt.split('\n')
    if len(out) <= (2 * cnt):
        return fmt
    return ''.join(out[:cnt]) + ' ...,' + ''.join(out[-cnt:])

def ichunk(iterable: Iterable[Any], n: int) -> Iterable[List[Any]]:
    """Yield n sized chunks, with tail chunk truncated.

    Args:
      iterable: sequence of values to be chunked
      n: size of chunks

    Returns:
      Yield an iterable of n sized chunks as lists

    >>> list(ichunk([], 3))
    []
    >>> list(ichunk([1], 3))
    [[1]]
    >>> list(ichunk(range(3), 3))
    [[0, 1, 2]]
    >>> list(ichunk(range(10), 3))
    [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
    """
    _iter = iter(iterable)
    while True:
        out = []
        for _ in range(n):
            try:
                out.append(next(_iter))
            except StopIteration:
                if out:
                    yield out
                return
        yield out



from typing import Generator

from jsonschema import Draft7Validator
from jsonschema.exceptions import ValidationError

class Validator:
    """A schema validator helper."""

    def __init__(self, schema: dict, config: dict):
        """Initialize Validator.

        Args:
          schema: validation schema
          config: config object to be validated
        """
        self.inst = Draft7Validator(schema)
        self.config = config

    def is_valid(self) -> bool:
        """Get boolean specifying if the config is valid.

        Returns:
          True if config is valid otherwise False
        """
        return self.inst.is_valid(self.config)

    def errors(self) -> Generator[ValidationError, None, None]:
        """Get iterator of errors from config validation.

        Returns:
          An iterator of all errors
        """
        return self.inst.iter_errors(self.config)

    def fmt(self, _e: ValidationError) -> str:
        """Format of an error showing the path and message.

        Args:
          _e: validation error

        Returns:
          A formatted string of the validation error
        """
        return f"Path({', '.join([str(_p) for _p in _e.path])}): {_e.message}"



from datetime import datetime
import logging
from typing import Callable, Dict, List, Iterable, Optional, Union
import pytz

from googleads import ad_manager

from .yaml_date import date_from_string
from .utils import values_from_bucket, ichunk, load_file, load_package_file, read_package_file

logging.basicConfig()

VERBOSE1: int = logging.INFO - 1
VERBOSE2: int = logging.INFO - 2

class Config:

    def __init__(self):
        self._schema = None
        self._cpm_names = None
        self._app = None
        self._start_time = datetime.now()
        self.set_logger()

    def isLoggingEnabled(self, level: int) -> bool:
        return self._logger.getEffectiveLevel() <= level

    def set_logger(self) -> None:
        self._logger = logging.getLogger(__package__)
        self._logger.setLevel(logging.INFO)
        logging.addLevelName(VERBOSE1, 'VERBOSE1')
        logging.addLevelName(VERBOSE2, 'VERBOSE2')

    def getLogger(self, name: str) -> logging.Logger:
        return self._logger.getChild(name.split('.')[-1])

    def set_log_level(self) -> None:
        if self.cli.get('verbose'):
            self._logger.setLevel(logging.INFO - len(self.cli['verbose']))
        if self.cli.get('quiet'):
            self._logger.setLevel(logging.WARNING)

    @property
    def app(self) -> dict:
        if self._app is None:
            self._app = self.settings_obj()
        return self._app

    @property
    def cli(self) -> dict:
        return self._cli

    @cli.setter
    def cli(self, obj) -> None:
        self._cli = obj
        self._client = None
        self.set_log_level()

    @property
    def client(self) -> Optional[ad_manager.AdManagerClient]:
        if self._client is None:
            self._client = self._client_factory(self.network_code, self.cli['private_key_file'])
        return self._client

    @property
    def user(self) -> dict:
        return self._user

    def set_client_factory(self, factory: Callable) -> None:
        self._client = None
        self._client_factory = factory

    def set_user_configfile(self, filename: str) -> None:
        self._user = load_file(filename)
        self._client = None
        self._cpm_names = None

    @property
    def start_time(self) -> datetime:
        return self._start_time

    @property
    def network_code(self) -> int:
        return self.cli['network_code'] or self.user.get('publisher', {}).get('network_code')

    @property
    def network_name(self) -> str:
        return self.cli['network_name'] or self.user.get('publisher', {}).get('network_name')

    @property
    def schema(self) -> dict:
        if self._schema is None:
            self._schema = load_file(self.cli['schema']) if self.cli.get('schema') else \
              load_package_file('schema.yml')
        return self._schema

    def bidder_codes(self) -> List[str]:
        if self.cli['single_order']:
            return [self.app['prebid']['bidders']['single_order']['code']]
        return self.cli['bidder_code']

    def media_types(self) -> List[str]:
        return [m_ for m_ in ('video', 'banner') if self.user['creative'].get(m_)]

    def targeting_bidder_key_config(self) -> Dict:
        return self.user.get('targeting', {}).get('bidder', {})

    def custom_targeting_key_values(self) -> List[Dict]:
        return [
            dict(
                name=_c['name'],
                values=set(_c['values']),
                operator=_c.get('operator', 'IS'),
                reportableType=_c.get('reportableType', 'OFF'),
            ) \
            for _c in self.user.get('targeting', {}).get('custom', [])
        ]

    def cpm_buckets(self) -> List[Dict[str, float]]:
        _type = self.user['rate']['granularity']['type']
        if _type == "custom":
            return self.user['rate']['granularity']['custom']
        return self.app['prebid']['price_granularity'][_type]

    def cpm_names(self) -> List[str]:
        if self._cpm_names is None:
            values = set()
            for bucket in self.cpm_buckets():
                values.update(values_from_bucket(bucket))
            self._cpm_names = ['%.2f' % v_ for v_ in sorted(values)]
        if self.cli['test_run']:
            return self._cpm_names[:self.app['mgr']['test_run']['line_item_limit']]
        return self._cpm_names

    def cpm_names_batched(self) -> Iterable[List[str]]:
        return ichunk(self.cpm_names(), self.app['googleads']['line_items']['max_per_order'])

    def micro_amount(self, cpm: Union[str, float]) -> int:
        return int(float(cpm) * self.app['googleads']['line_items']['micro_cent_factor'])

    def template_src(self) -> str:
        if self.cli.get('template'):
            with open(self.cli['template']) as fp:
                return fp.read()
        return read_package_file('line_item_template.yml')

    def settings_obj(self) -> dict:
        if self.cli.get('settings'):
            return load_file(self.cli['settings'])
        return load_package_file('settings.yml')

    def pre_create(self) -> None:
        li_ = self.user['line_item']
        is_standard = li_['item_type'].upper() == "STANDARD"
        is_sponsorship = li_['item_type'].upper() == "SPONSORSHIP"
        end_str = li_.get('end_datetime')
        start_str = li_.get('start_datetime')
        fmt = self.app['mgr']['date_fmt']
        vcpm = self.user['rate'].get('vcpm')

        if self.user['creative'].get('video'):
            for i_ in ('max_duration', ):
                max_duration = self.user['creative']['video'].setdefault(
                    i_, self.app['prebid']['creative']['video'][i_])
            _ = self.user['creative']['video'].setdefault('duration', max_duration)

        if vcpm and not is_standard:
            raise ValueError("Specifying 'vcpm' requires using line item type 'standard'")

        try:
            tz_str = li_.get('timezone', self.app['mgr']['timezone'])
            _ = pytz.timezone(tz_str)
        except pytz.exceptions.UnknownTimeZoneError as e:
            raise ValueError(f'Unknown Time Zone, {e}') from e

        for i_ in ('line_item', 'order'):
            self.user[i_]['name'] = ''.join(['{{ run_mode }}', self.user[i_]['name']])

        li_.update(dict(
            start_dt=date_from_string(start_str, fmt, tz_str) if start_str else "IMMEDIATELY",
            start_dt_type="USE_START_DATE_TIME" if start_str else "IMMEDIATELY",
            end_dt=date_from_string(end_str, fmt, tz_str),
            unlimited_end_dt=not end_str,
        ))

        if not is_sponsorship:
            li_.update({'goal': dict(
                goalType="NONE",
            )})

        if vcpm:
            li_.update({'goal': dict(
                goalType="LIFETIME",
                unitType="VIEWABLE_IMPRESSIONS",
                units=vcpm,
            )})

        self.user['rate'].update(dict(
            cost_type="VCPM" if vcpm else "CPM"
        ))

config = Config()



"""Console script for line_item_manager."""
from functools import partial
import json
import pkg_resources
import sys

import click
from googleads.errors import GoogleAdsError
import yaml

from . import version as VERSION
from .config import config
from .exceptions import ResourceNotActive, ResourceNotFound
from .gam_config import GAMConfig
from .gam_operations import client as gam_client
from .prebid import prebid, PrebidBidder
from .validate import Validator

click.option = partial(click.option, show_default=True)

logger = config.getLogger(__name__)

@click.group(invoke_without_command=True)
@click.option('--version', is_flag=True, help='Print version information and exit.')
@click.pass_context
def cli(ctx: click.core.Context, version: bool) -> None:
    if version:
        print(f'line-item-manager version {VERSION}')
        return
    if not ctx.invoked_subcommand:
        click.echo(cli.get_help(ctx))

@cli.command()
@click.argument('configfile', type=click.Path(exists=True))
@click.option('--network-code', type=int,
              help='GAM network code, must reconcile with the network name.')
@click.option('--network-name', help='GAM network name, must reconcile with the network code.')
@click.option('--private-key-file', '-k', required=True, default='gam_creds.json',
              type=click.Path(exists=True), help='Path to json GAM credentials file.')
@click.option('--template',
              type=click.Path(exists=True), help='Advanced users: path to custom line item template. ' \
              'Use "line_item_manager show template" to see the package default')
@click.option('--settings',
              type=click.Path(exists=True), help='Advanced users: path to settings file. ' \
              'Use "line_item_manager show settings" to see the package default')
@click.option('--schema',
              type=click.Path(exists=True), help='Advanced users: path to schema file. ' \
              'Use "line_item_manager show schema" to see the package default')
@click.option('--single-order', '-s', is_flag=True,
              help='Create a single set of orders instead of orders per bidder.')
@click.option('--bidder-code', '-b', multiple=True,
              help='Bidder code, may be used multiple times.')
@click.option('--test-run', '-t', is_flag=True,
              help='Create a limited number of line_items for testing and review in GAM. ' \
              'Order and line item names are prepended with \'Test: \'.')
@click.option('--dry-run', '-n', is_flag=True,
              help='Print commands that would be executed, but do not execute them.')
@click.option('--quiet', '-q', is_flag=True, help='Logging is limited to warnings and errors.')
@click.option('--verbose', '-v', multiple=True, is_flag=True,
              help='Verbose logging, use multiple times to increase verbosity.')
@click.option('--skip-auto-archive', is_flag=True,
              help='Upon failure or interruption, do NOT auto-archive already created orders.')
@click.pass_context
def create(ctx: click.core.Context, configfile: str, **kwargs):
    """Create line items"""
    config.cli = kwargs

    try:
        config.set_user_configfile(configfile)
    except yaml.YAMLError as e:
        raise click.UsageError(f'Check your configfile. {e}', ctx=ctx)

    gam = GAMConfig()

    # validate CLI options
    if not kwargs['single_order'] and not kwargs['bidder_code']:
        raise click.UsageError(
            'You must use --single-order or provide at least one --bidder-code', ctx=ctx)

    if kwargs['single_order'] and kwargs['bidder_code']:
        raise click.UsageError(
            'Use of --single-order and --bidder-code is ambiguous and not allowed.', ctx=ctx)

    if not config.network_code:
        raise click.UsageError(
            'Network code must be provided as an option or set in the config file', ctx=ctx)

    if not config.network_name:
        raise click.UsageError(
            'Network name must be provided as an option or set in the config file', ctx=ctx)

    for bidder_code in kwargs['bidder_code']:
        if not bidder_code in prebid.bidders:
            raise click.UsageError(f'Bidder code \'{bidder_code}\' is not recognized', ctx=ctx)

    # validate GAM client
    config.set_client_factory(gam_client)
    try:
        config.client
    except json.decoder.JSONDecodeError as e:
        raise click.UsageError(
            'Check your private key file. File is not formatted properly as json', ctx=ctx)
    except ValueError as e:
        raise click.UsageError(f'Check your private key file. {e.args[0]}', ctx=ctx)
    except Exception as e:
        raise click.UsageError('Check your private key file. Not able to successfully ' \
                               'access your service account', ctx=ctx)

    # validate GAM network access
    try:
        if not gam.network['displayName'] == config.network_name:
            raise click.UsageError(f"Network name found \'{gam.network['displayName']}\' " \
                                   f"does not match provided \'{config.network_name}\'", ctx=ctx)
    except GoogleAdsError as _e:
        logger.error(f'GoogleAdsError, {_e}')
        raise click.UsageError(
            'Not able to successfully access your service account.  ' \
            'Check WARNING/ERROR messages above and consider \nupgrading to latest version.', ctx=ctx)

    # validate user configuration
    user_cfg = Validator(config.schema, config.user)
    if not user_cfg.is_valid():
        err_str = '\n'.join([f'  - {user_cfg.fmt(_e)}' for _e in user_cfg.errors()])
        raise click.UsageError(f'Check your configfile for the following validation ' \
                               f'errors:\n{err_str}', ctx=ctx)
    try:
        PrebidBidder.validate_override_map(config.user.get('bidder_key_map'))
    except ValueError as e:
        raise click.UsageError(f'{e}', ctx=ctx)

    # pre-create line items config
    try:
        config.pre_create()
    except ValueError as e:
        raise click.UsageError(f'{e}', ctx=ctx)

    # create line items
    try:
        gam.create_line_items()
        gam.success = True
    except ResourceNotActive as _e:
        logger.error('Resource is not active:\n  - %s', _e)
    except ResourceNotFound as _e:
        logger.error('Not able to find the following resource:\n  - %s', _e)
    except GoogleAdsError as _e:
        logger.error('Google Ads Error, %s', _e)
    except ValueError as _e:
        logger.error('Unexpected result, %s', _e)
    except KeyboardInterrupt:
        logger.warning('User Interrupt')
    finally:
        try:
            gam.cleanup()
        except GoogleAdsError as _e:
            logger.error('Cleanup: Google Ads Error, %s', _e)

def show_resource(filename: str) -> None:
    rsrc_name = pkg_resources.resource_filename('line_item_manager', filename) # type: ignore[misc]
    with open(rsrc_name) as fp:
        print(fp.read())

@cli.command()
@click.argument('resource', type=click.Choice(['config', 'bidders', 'template',
                                               'settings', 'schema']))
def show(resource: str) -> None:
    """Show resources"""
    if resource == 'config':
        show_resource('conf.d/line_item_manager.yml')
    if resource == 'template':
        show_resource('conf.d/line_item_template.yml')
    if resource == 'settings':
        show_resource('conf.d/settings.yml')
    if resource == 'schema':
        show_resource('conf.d/schema.yml')
    if resource == 'bidders':
        print("%-25s%s" % ('Code', 'Name'))
        print("%-25s%s" % ('----', '----'))
        for row in sorted(prebid.bidders.values(), key=lambda x: x['bidder-code']):
            print("%-25s%s" % (row['bidder-code'], row['bidder-name']))

def main():
    cli() # pylint: disable=no-value-for-parameter

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



from pprint import pformat
from typing import Any, List, Optional, Tuple

from googleads import ad_manager
from googleads.common import ZeepServiceProxy
import yaml

from .config import config, VERBOSE2
from line_item_manager.utils import load_package_file

logger = config.getLogger('operations')

_CREATE_LOG_LINE = 'Create: "%s" w/ Rec(s):\n"%s"'
_QUERY_LOG_LINE = 'Service: "%s" Method: "%s" Params:\n"%s"'
_RESULTS_LOG_LINE = 'Service: "%s" Method: "%s" Results:\n"%s"'

def client(network_code: int, key_file: str) -> ad_manager.AdManagerClient:
    _cfg = load_package_file('googleads.yaml')
    _cfg['ad_manager']['network_code'] = network_code
    _cfg['ad_manager']['path_to_private_key_file'] = key_file
    return ad_manager.AdManagerClient.LoadFromString(yaml.dump(_cfg))

class GAMOperations:
    service: str = ''
    method: str = ''
    create_method: str = ''
    query_fields: Optional[Tuple[str, ...]] = None
    create_fields: Optional[Tuple[str, ...]] = None
    log_fields: Optional[Tuple[str, ...]] = None
    use_statement = True

    def __init__(self, **kwargs):
        self.params = kwargs
        self.query_params = {k:kwargs[k] for k in self.query_fields if k in kwargs} \
          if self.query_fields else self.params
        self.create_params = {k:kwargs[k] for k in self.create_fields if k in kwargs} \
          if self.create_fields else self.params

    def create(self, atts: List[dict], validate: bool=False, verbose: bool=True) -> List[dict]:
        logger.log(VERBOSE2, _CREATE_LOG_LINE, type(self).__name__, pformat(self.log_recs(atts)))
        results = self.dry_run_recs(atts) if self.dry_run else \
          getattr(self.svc(), self.create_method)(atts)
        if verbose:
            logger.log(VERBOSE2, _RESULTS_LOG_LINE, self.service, self.method, pformat(results))
        if validate:
            self.validate(atts, results)
        return results

    def fetch(self, one: bool=False, create: bool=False, recs: List[dict]=None,
              validate: bool=False) -> List[dict]:
        logger.log(VERBOSE2, _QUERY_LOG_LINE, self.service, self.method, pformat(self.query_params))
        results = self._results(one=one)
        if create and recs:
            current = {self.check(r_) for r_ in results}
            new_recs = [r_ for r_ in recs if self.check(r_) not in current]
            if new_recs:
                results += self.create(new_recs, verbose=False)
        elif create and not results:
            results = self.create([self.create_params], verbose=False)
        logger.log(VERBOSE2, _RESULTS_LOG_LINE, self.service, self.method, pformat(results))
        if validate:
            self.validate(recs, results)
        return results

    def fetchone(self, **kwargs) -> dict:
        recs = self.fetch(one=True, **kwargs)
        return recs[0] if recs else {}

    def _results(self, one: bool=False) -> List[dict]:
        if not self.use_statement:
            return [getattr(self.svc(), self.method)()]
        _stm = self.statement()
        results = []
        while True:
            response = getattr(self.svc(), self.method)(_stm.ToStatement())
            if not ("results" in response and response["results"]):
                break
            for result in response["results"]:
                if one:
                    return [result]
                results.append(result)
            if len(results) < _stm.limit:
                break
            _stm.offset += _stm.limit
        return results

    def statement(self) -> ad_manager.StatementBuilder:
        _stm = ad_manager.StatementBuilder(version=self.version)
        if self.query_params:
            _stm.Where(' AND '.join([f"{k} = :{k}" for k in self.query_params]))
            _ = [_stm.WithBindVariable(k, v) for k, v in self.query_params.items()]
        return _stm

    def svc(self) -> ZeepServiceProxy:
        return self.client.GetService(self.service, version=self.version) # type: ignore[union-attr]

    def log_recs(self, recs: List[dict]) -> List[dict]:
        if self.log_fields:
            return [{f_:r_[f_] for f_ in self.log_fields if f_ in r_} for r_ in recs]
        return recs

    @property
    def client(self) -> Optional[ad_manager.AdManagerClient]:
        raise NotImplementedError

    @property
    def version(self) -> str:
        raise NotImplementedError

    @property
    def dry_run(self) -> bool:
        raise NotImplementedError

    def dry_run_recs(self, recs: List[dict]) -> List[dict]:
        raise NotImplementedError

    def check(self, rec: dict) -> Any:
        raise NotImplementedError

    def validate(self, recs, results):
        raise NotImplementedError



import re
import yaml

from jinja2 import Template as J2Template

from .config import config
from .prebid import PrebidBidder

logger = config.getLogger(__name__)

JINJA_PATTERN = re.compile(r'{{\W*(\w*)\W*}}')

def render_src(src: str, **kwargs) -> dict:
    """Get object from jinja rendered yaml removing non-word chars from variable references.

    Args:
      src: templated source string
      kwargs: key-value pairs referenced in template

    Returns:
      A dict of the jinja rendered src
    """
    clean_src = JINJA_PATTERN.sub(r'{{ \1 }}', src)
    return yaml.safe_load(J2Template(clean_src).render(**kwargs))

def render_cfg(objname: str, bidder: PrebidBidder, media_type: str=None,
               cpm: str=None, cpm_min: str=None, cpm_max: str=None) -> dict:
    """Get jinja rendered object of a top level user config object.

    Args:
      objname: top level object name in user provided config
      bidder: a bidder instance
      media_type: media type value
      cpm: cpm value
      cpm_min: cpm minimum value
      cpm_max: cpm maximum value


    Returns:
      A dict of the jinja rendered src
    """
    params = dict(
        time=config.start_time.strftime("%m/%d/%Y-%H:%M:%S"),
        run_mode='Test: ' if config.cli['test_run'] else '',
        bidder_code=bidder.codestr,
        bidder_name=bidder.name,
    )
    params.update(bidder.params)
    if media_type:
        params['media_type'] = media_type
    if cpm:
        params['cpm'] = cpm
    if cpm_min:
        params['cpm_min'] = cpm_min
    if cpm_max:
        params['cpm_max'] = cpm_max
    return render_src(yaml.safe_dump(config.user[objname]), **params)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages
import os.path

def get_txt(filename):
    with open(filename) as txt_file:
        return txt_file.read()

readme = get_txt('README.rst') if os.path.isfile('README.rst') else ''
history = get_txt('HISTORY.rst') if os.path.isfile('HISTORY.rst') else ''

requirements = [
    'Click==8.1.7',
    'googleads==39.0.0',
    'jinja2==3.0.3',
    'jsonschema==4.23.0',
    'PyYAML==6.0.1',
    'retrying==1.3.3',
    'tqdm==4.56.0',
]

package_data = [
    'conf.d/*.yml',
    'conf.d/*.yaml',
]

release_requirements = [
    'bump2version>=1',
    'twine>=5',
]

setup_requirements = []

test_requirements = [
    'flake8==3.8.4',
    'mock==4.0.2',
    'pytest==7.0.0',
    'pytest-cov==3.0.0',
    'pytest-runner==5.3.1',
 ]

setup(
    author="the prebid contributors",
    author_email='info@prebid.org',
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    description="Create and manage line items.",
    entry_points={
        'console_scripts': [
            'line_item_manager=line_item_manager.cli:main',
        ],
    },
    extras_require={
        'release': release_requirements,
        'test': test_requirements,
    },
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='line-item-manager',
    name='line-item-manager',
    packages=find_packages(include=['line_item_manager', 'line_item_manager.*']),
    package_dir={'line_item_manager': 'line_item_manager'},
    package_data={'line_item_manager': package_data},
    setup_requires=setup_requirements,
    test_suite='tests',
    url='https://github.com/prebid/line-item-manager',
    version='0.2.13',
    zip_safe=False,
)



import certifi
import csv
from typing import  Any, Dict, Optional
from urllib import request
import ssl

from .utils import load_package_file

SETTINGS: dict = load_package_file('settings.yml')['prebid']
BIDDERS: dict = SETTINGS['bidders']

class Prebid:
    """Prebid specific."""

    def __init__(self):
        """Initialize Prebid."""
        self._bidders = None

    @property
    def bidders(self) -> Dict[str, Any]:
        """Read remote bidders CSV metadata file.

        Returns:
          A dict keyed by bidder code
        """
        if self._bidders is None:
            context = ssl.create_default_context(cafile=certifi.where())
            reader = csv.DictReader([l.decode('utf-8') for l in request.urlopen(
                BIDDERS['data'], context=context).readlines()], skipinitialspace=True)
            self._bidders = {row['bidder-code']:row for row in reader}
        return self._bidders
prebid = Prebid()

class PrebidBidder:
    """Prebid Bidder."""

    def __init__(self, code: str, override_map: Dict[str, str]=None,
                 single_order: bool=False):
        """Initialize PrebidBidder.

        Args:
          code: bidder code
          override_map: mapping of bidder keys to override values
          single_order: True if single order
        """
        self.code = code
        self.name = BIDDERS['single_order']['name'] if single_order \
          else prebid.bidders[code]['bidder-name']
        self.override_map = override_map or {}
        self.single_order = single_order
        self._params: Optional[Dict[str, str]] = None

    @property
    def codestr(self) -> str:
        """String representation.

        Returns:
          Empty string if single order else code
        """
        return '' if self.single_order else self.code

    @staticmethod
    def validate_override_map(key_map: Optional[dict]) -> None:
        """Validate an override map checking for valid bidders and keys.

        Args:
          key_map: override map to validate
        """
        if not key_map:
            return
        valid_properties = set(prebid.bidders.keys())
        valid_properties.add(BIDDERS['targeting_key'])
        if not set(key_map.keys()).issubset(valid_properties):
            raise ValueError(f"'bidder_key_map' properties, " \
                             f"{set(key_map.keys())} must be valid bidder codes")
        valid_keys = set(BIDDERS['keys'])
        for k, v in key_map.items():
            if not set(v.keys()).issubset(valid_keys):
                raise ValueError(f"'bidder_key_map' properties, {set(v.keys())}, " \
                                 f"for '{k}' must be valid bidder keys")

    @property
    def params(self) -> Dict[str, str]:
        """Attribute of bidder specific parameters keyed by prefixes.

        Returns:
          Dict of bidder specific parameters
        """
        if self._params is None:
            self._params = \
              {k:self.override_map.get(k, self.fmt_bidder_key(k)) for k in BIDDERS['keys']}
        return self._params

    @property
    def targeting_key(self) -> str:
        """Get bidder specific targeting key.

        Returns:
          Bidder targeting key
        """
        return self.params[BIDDERS['targeting_key']]

    def fmt_bidder_key(self, prefix: str) -> str:
        """Get formatted bidder key with character limit.

        Args:
          prefix: bidder parameters prefix

        Returns:
          Formatted bidder key
        """
        if not self.codestr:
            return prefix
        return f'{prefix}_{self.codestr}'[:BIDDERS['key_char_limit']]
